
//  PARSER.CC  Ver. 1.0
//  Program to extract Nodal equations from Spice Netlist.  Ed Chan

#include "parser.h"

double stripString( char *stringIn );

main( int argc, char *argv[] ) {
  ifstream inFile;
  ofstream outFile;
  NodeHead nodeList;
  CompHead compList;
  ModelHead modelList;

  // Buffers used in parsing:
  char inName[NameLength], outName[NameLength], buf[BufLength], 
    buf1[BufLength], buf2[BufLength], buf3[BufLength], *bufPtr,
    *charPtr1, *charPtr2;
  int intBuf1, intBuf2, intBuf3, intBuf4, datum;
  double douBuf1, douBuf2, douBuf3, douBuf4;
  CompType typeBuf;
  Component *compPtr, *compPtr1, *compPtr2;
  Node *nodePtr, *nodePtr1, *nodePtr2;
  Model* modelPtr;
  TranType TtypeBuf;


  //  setting up the input and output files:
  if (argc!=2){
    cout << "Please enter the input Spice Netlist: <QUIT to exit>" << endl;
    cin >> inName;
    if( !strcmp( inName, "QUIT" ) ){
      cerr << "Program Exited Abnormally!" << endl;
	  exit(0);
    } else {
      inFile.open( inName, ios::in );
      while( !inFile ){
	cerr << inName << " is an invalid input." << endl
	     << "Please enter the input Spice Netlist: <QUIT to exit>" << endl;
	cin >> inName;
	inFile.open( inName, ios::in );
      }
    }
  }
  else{
	strcpy( inName, argv[1] );
    inFile.open( inName, ios::in );
    while( !inFile ){
      cerr << inName << " is an invalid input." << endl
	   << "Please enter the input Spice Netlist: <QUIT to exit>" << endl;
      cin >> inName;
	  inFile.open( inName, ios::in );
    }
  }

  strcpy( outName, inName );
  strtok( outName, "." );
  strcat( outName, ".Pout" );
  outFile.open( outName, ios::out );
  cout << "Output saved to file: " << outName << endl;
  outFile << "Circuit from file: " << inName << " using Parser." << endl;


  // parsing of netlist to create linked list of models (remember to reset the fstream)
  inFile.getline(buf, BufLength);       // first line of netlist is discarded
  inFile.getline(buf, BufLength);
  while( inFile ){
    strcpy( buf1, buf );
    if( !strcmp( strtok( buf1, " " ), ".model" ) ){
      strcpy( buf2, strtok( NULL, " " ) );
      charPtr1 = strtok( NULL, " " );
      if( !strcmp( charPtr1, "PNP" ) )
	TtypeBuf = PNP;
      else if( !strcmp( charPtr1, "NPN" ) )
	TtypeBuf = NPN;
      else if( !strcmp( charPtr1, "NMOS" ) )
	TtypeBuf = NMOS;
      else if( !strcmp( charPtr1, "PMOS" ) )
	TtypeBuf = PMOS;
    
      charPtr1 = strtok( NULL, " " );
      while ( charPtr1 != NULL ){
	strcpy( buf3, "" );
	if( (charPtr1[0] == 'I') && (charPtr1[1] == 'S') && (charPtr1[2] == '=') ){
      	  douBuf1 = stripString( charPtr1 );
	}
	if( (charPtr1[0] == 'B') && (charPtr1[1] == 'F') && (charPtr1[2] == '=') ){
	  douBuf2 = stripString( charPtr1 );
	}
	if( (charPtr1[0] == 'B') && (charPtr1[1] == 'R') && (charPtr1[2] == '=') ){
	  douBuf3 = stripString( charPtr1 );
	}
	if( (charPtr1[0] == 'T') && (charPtr1[1] == 'E') && (charPtr1[4] == '=') ){
	  douBuf4 = stripString( charPtr1 );
	}
	charPtr1 = strtok( NULL, " " );
      }
      modelPtr = new Model( buf2, TtypeBuf, douBuf1, douBuf2, douBuf3, douBuf4 );
      modelList.addModel( modelPtr );
    }
    inFile.getline(buf, BufLength);
  }
  inFile.close();
  inFile.open( inName, ios::in );


  //  starting of parsing by creating linked list of components
  inFile.getline(buf, BufLength);       // first line of netlist is discarded
  inFile.getline(buf, BufLength);
  while( inFile ){
    if( isalpha( *buf ) ){
      
      //  EDIT THIS SECTION IF NEW COMPONENTS ARE ADDED!!!
      //  we could do some rearranging in this section to catch each type in order.
      switch( *buf ) {
      case 'v':
      case 'V':
	typeBuf = VSource;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	douBuf1 = atof( strtok( NULL, " " ) );
	compPtr = new Component( typeBuf, douBuf1, NA,  intBuf1, intBuf2, NA, NA, NULL );
	compList.addComp( compPtr );
	break;
      case 'q':
      case 'Q':
	typeBuf = BJT;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	intBuf3 = atoi( strtok( NULL, " " ) );
	compPtr = new Component( typeBuf, NA, NA, intBuf1, intBuf2, intBuf3, NA, 
				 modelList.getModel( strtok( NULL, " " ) ) );
	compList.addComp( compPtr );
	break;
      case 'm':
      case 'M':
	typeBuf = MOSFET;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	intBuf3 = atoi( strtok( NULL, " " ) );
	intBuf4 = atoi( strtok( NULL, " " ) );	
	compPtr = new Component( typeBuf, NA, NA, intBuf1, intBuf2, intBuf3, intBuf4, 
				 modelList.getModel( strtok( NULL, " " ) ) );
	compList.addComp( compPtr );
	break;
      case 'r':
      case 'R':
	typeBuf = Resistor;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	douBuf1 = atof( strtok( NULL, " " ) );
	compPtr = new Component( typeBuf, douBuf1, NA, intBuf1, intBuf2, NA, NA, NULL );
	compList.addComp( compPtr );
	break;
      case 'd':
      case 'D':
	typeBuf = Diode;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	charPtr1 = strtok( NULL, " " );
	while( charPtr1 != NULL ){
	  if( (charPtr1[0] == 'I') && (charPtr1[1] == 'S') && (charPtr1[2] == '=') ){
	    douBuf1 = stripString( charPtr1 );
	  }	    
	  if( (charPtr1[0] == 'T') && (charPtr1[1] == 'E') && (charPtr1[4] == '=') ){
	    douBuf2 = stripString( charPtr1 );
	  }
	  charPtr1 = strtok( NULL, " " );
	}
	compPtr = new Component( typeBuf, douBuf1, douBuf2, intBuf1, intBuf2, NA, NA, NULL );
	compList.addComp( compPtr );
	break;
      case 'c':
      case 'C':
	typeBuf = Capacitor;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	douBuf1 = atof( strtok( NULL, " " ) );
	compPtr = new Component( typeBuf, douBuf1, NA, intBuf1, intBuf2, NA, NA, NULL);
	compList.addComp( compPtr );
	break;
      case 'l':
      case 'L':
	typeBuf = Inductor;
	strtok( buf, " " );
	intBuf1 = atoi( strtok( NULL, " " ) );
	intBuf2 = atoi( strtok( NULL, " " ) );
	douBuf1 = atof( strtok( NULL, " " ) );
	compPtr = new Component( typeBuf, douBuf1, NA, intBuf1, intBuf2, NA, NA, NULL);
	compList.addComp( compPtr );
	break;
      };
    }
    inFile.getline(buf, BufLength);
  }


//  Now the components are created and it is time to set up the list of nodes.
//  we should actually use second connector of first Source as the first Node (Datum)
  compPtr1 = compList.getComp(0);
  while( compPtr1 != NULL ){
	for ( int b=0; b < 3; b++ ){
	  if( (!compPtr1->isCon( b )) && ( compPtr1->getConVal( b ) != NA ) ){
	intBuf1 = compPtr1->getConVal( b );
	nodePtr1 = nodeList.addNode();
	compPtr1->connect( b, nodePtr1 );
	nodePtr1->connect( b, compPtr1 );
	
	// now search and connect all other appropriate connectors to this node.
	// error checking should be added to prevent duplicated, or skipped connectors.
	compPtr2 = compPtr1->getNext();
	while( compPtr2 != NULL ){
	  for( int c=0; c < 3; c++ ){
	    if( compPtr2->getConVal( c ) == intBuf1  ){
	      compPtr2->connect( c, nodePtr1 );
	      nodePtr1->connect( c, compPtr2 );
	      break;
	    }
	  }  
	  compPtr2 = compPtr2->getNext();
	}
      }
    }
    compPtr1 = compPtr1->getNext();
  }

  
//  At this point, we are done creating a representation of the circuit in memory
//  now, we need to call each node to create and output its nodal equation.
//  Each node will call the components attached for the individual contributions to the
//  nodal equation.



  //  Loop to find the datum
  nodePtr = nodeList.getNode(0);
  nodePtr1 = nodePtr->getNext();
  while( nodePtr1 != NULL ){
    if( nodePtr1->getCount() > nodePtr->getCount() )
      nodePtr = nodePtr1;
    nodePtr1 = nodePtr1->getNext();
  }
  datum = nodePtr->getNum();
  
  // go down the nodal list and have components announce themselves

  nodePtr = nodeList.getNode(0);
  while( nodePtr != NULL ){
    if( nodePtr->getNum() != datum ){
      outFile << endl	<< "Node " << nodePtr->getNum() << " connections: " << endl;
      nodePtr->printComps( outFile, datum );
    }
    else{
      outFile << endl << "Node " << nodePtr->getNum() << " is the Datum.  It has equation: " << endl;
      nodePtr->printComps( outFile, datum );
    }
    nodePtr = nodePtr->getNext();
  }
  
  // go down the component list and give modified nodal equations for all sources
  
  outFile << endl << "Modified nodal equations for Voltage Sources: " << endl;
  compPtr = compList.getComp(0);
  while( compPtr != NULL ){
    compPtr->specialPrint( outFile, datum );
    compPtr = compPtr->getNext();
  }
  
  cout << endl << "That's all folks!" << endl;
  
  
  return 0;
}




double stripString( char *stringIn ){
  char buf[BufLength], buf2[BufLength];
  strcpy( buf, stringIn );
  for( int a=0; buf[a] != '='; a++ ){};
  a++;
  for( int b=0; buf[a] != '\0'; b++, a++ )
    buf2[b] = buf[a];
  buf2[b] = '\0';
  return atof( buf2 );
};







