//  PARSER.H
//  Header for Parser program.

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <string.h>

//  declarations:
enum CompType { MOSFET, BJT, VSource, ISource, Inductor, 
		  Resistor, Diode, Capacitor };
enum TranType { NMOS, PMOS, NPN, PNP };
enum Flag { UNSET, SET };
enum Boolean { FALSE, TRUE };
int mCount=0, bCount=0, vSCount=0, iSCount=0, rCount=0, iCount=0, dCount=0, cCount=0;
const int NameLength=80, BufLength=300, NA=-1;

class Component;
class ComponentHead;
class Node;
class NodeHead;
class Model;
class ModelHead;

struct Connectors{
	Flag flag;
	Node* node;
	int conNum;
};

struct Connections{
	Connections* next;
	Component* comp;
	int conNum;
};


class Component{
public:
  Component( CompType typeIn, double valueIn, double tempIn, int con0In, 
	     int con1In, int con2In, int con3In, Model* modelIn );
  ~Component();
  CompType getType();
  int getNum();
  Component* getNext();
  void connect( int conNum, Node* nodeIn );
  double getVal();
  void setNext( Component* nextIn );
  void setNum( int numIn );
  int getConVal( int conNum );
  Boolean isCon( int conNum );
  void print( int nodeNum, ofstream &outFile, int datum );
  void specialPrint( ofstream &outFile, int datum );
  Node* getNode( int conNum );
  int Component::getNodeNum( int conNum );
private:
  Component* next;
  CompType type;
  Connectors con0, con1, con2, con3;
  int compNum;
  double value, temp;
  Model* model;
};

class Node{
public:
  Node(int num);
  ~Node();
  int getNum();
  int getCount();
  void connect( int conNumIn, Component* compIn );
  Node* getNext();
  void printComps( ofstream &outFile, int datum );
  void setNext( Node* nodeIn );
private:
  Node* next;
  int nodeNum, conCount;
  Connections* conList;
};

class NodeHead {
public:
  NodeHead();
  ~NodeHead();
  Node* addNode();
  int getCount();
  Node* getNode(int nodeNum);
private:
  Node* nodeList;
  int nodeCount;
};

class CompHead {
public:
  CompHead();
  ~CompHead();
  void addComp( Component* component );
  int getCount( CompType type );
  Component* getComp( int compNum );
private:
  Component* compList;
  int sCount, rCount, tCount, dCount, cCount;
};

class Model {
public:
  Model( char* nameIn, TranType typeIn, double isIn, double bfIn, 
	 double brIn, double tempIn );
  char* getName();
  TranType getType();
  double getIs();
  double getBf();
  double getBr();
  double getTemp();
  void setNext( Model* nextIn );
  Model* getNext();
private:
  char name[NameLength];
  double is, bf, br, temp;
  Model* next;
  TranType type;
};

class ModelHead{
public:
  ModelHead();
  void addModel( Model* modelIn );
  Model* getModel ( char* nameIn );
private:
  Model* modelList;
};



Component::Component( CompType typeIn, double valueIn=NA, double tempIn=NA, int con0In=NA, 
		     int con1In=NA, int con2In=NA, int con3In=NA, Model* modelIn=NULL ){
	type = typeIn;
	con0.conNum = con0In;
	con1.conNum = con1In;
	con2.conNum = con2In;
	con3.conNum = con3In;
	con0.flag = UNSET;
	con1.flag = UNSET;
	con2.flag = UNSET;
	con3.flag = UNSET;
	value = valueIn;
	temp = tempIn;
	next = NULL;
	model = modelIn;
	}

Component::~Component(){};

void Component::connect( int conNum, Node* nodeIn ){
	if( conNum == 0 ){
		con0.node = nodeIn;
		con0.flag = SET;
		}
	if( conNum == 1 ){
		con1.node = nodeIn;
		con1.flag = SET;
		}
	if( conNum == 2 ){
		con2.node = nodeIn;
		con2.flag = SET;
		}
	if( conNum == 3 ){
		con3.node = nodeIn;
		con3.flag = SET;
		}
	}

CompType Component::getType(){
	return type;
	}

int Component::getNum(){
	return compNum;
	}

Component* Component::getNext(){
	return next;
	}

double Component::getVal(){
	return value;
	}

void Component::setNext( Component* nextIn ){
	next = nextIn;
	}

void Component::setNum( int numIn ){
	compNum = numIn;
	}

int Component::getConVal( int conNum ){
	int rtVal;
	if( conNum == 0 )
		rtVal = con0.conNum;
	if( conNum == 1 )
		rtVal = con1.conNum;
	if( conNum == 2 )
		rtVal = con2.conNum;
	if( conNum == 3 )
		rtVal = con3.conNum;
	return rtVal;
	}

Boolean Component::isCon( int conNum ){
	Boolean rtVal;
	if( conNum == 0 )
		rtVal = ( con0.flag == SET )? TRUE: FALSE;
	if( conNum == 1 )
		rtVal = ( con1.flag == SET )? TRUE: FALSE;
	if( conNum == 2 )
		rtVal = ( con2.flag == SET )? TRUE: FALSE;
	if( conNum == 3 )
		rtVal = ( con3.flag == SET )? TRUE: FALSE;
	return rtVal;
	}

void Component::print( int nodeNum, ofstream &outFile, int datum ){
	switch( type ){
	case MOSFET:
	  outFile << "MOSFET" << getNum()
	    << " connected: Drain->Node" << getNodeNum(0)
	      << " Source->Node" << getNodeNum(1)
		<< " Gate->Node" << getNodeNum(2) 
		  << " Bulk->Node" << getNodeNum(3)<< " ." << endl;
	  
	  
	case BJT:
	  if( (con0.node->getNum() == nodeNum) && (model->getType() == NPN) ){
	    outFile << "(-" << model->getIs() << ")(e^(-n(";
	    if( con2.node->getNum() != datum )
	      outFile << "X" << con2.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	    outFile << " + ";
	    outFile << "(" << model->getIs()
	      << "/" << model->getBr() 
		<< ")(1+" << model->getBr()
		  << ")(e^(-n(";
	    if( con0.node->getNum() != datum )
	      outFile << "X" << con0.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	  }
	  
	  if( (con2.node->getNum() == nodeNum) && (model->getType() == NPN) ){
	    outFile << "(" << model->getIs()
	      << "/" << model->getBf() 
		<< ")(1+" << model->getBf()
		  << ")(e^(-n(";
	    if( con2.node->getNum() != datum )
	      outFile << "X" << con2.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	    outFile << " - ";
	    outFile << "(" << model->getIs() << ")(e^(-n(";
	    if( con0.node->getNum() != datum )
	      outFile << "X" << con0.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	  }
	  
	  if( (con1.node->getNum() == nodeNum) && (model->getType() == NPN) ){
	    outFile << "(" << model->getIs() << ")(e^(-n(";
	    if( con2.node->getNum() != datum )
	      outFile << "X" << con2.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	    outFile << " + ";
	    outFile << "(-" << model->getIs()
	      << "/" << model->getBr() 
		<< ")(1+" << model->getBr()
		  << ")(e^(-n(";
	    if( con0.node->getNum() != datum )
	      outFile << "X" << con0.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	    outFile << " + ";
	    outFile << "(-" << model->getIs()
	      << "/" << model->getBf() 
		<< ")(1+" << model->getBf()
		  << ")(e^(-n(";
	    if( con2.node->getNum() != datum )
	      outFile << "X" << con2.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	    outFile << " + ";
	    outFile << "(" << model->getIs() << ")(e^(-n(";
	    if( con0.node->getNum() != datum )
	      outFile << "X" << con0.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")) -1)";
	  }
	  break;
	  
	  
	  
	  
	  /*
	    case BJT:
	    if( (con0.node->getNum() == nodeNum) && (model->getType() == NPN) )
	    outFile << " (-" << model->getIs() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) + (" << model->getIs() << "/"
	    << model->getBr() << ")(1+" << model->getBr() << ")(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    if( (con2.node->getNum() == nodeNum) && (model->getType() == NPN) )
	    outFile << " (" << model->getIs() << "/" << model->getBf() << ")(1+" 
	    << model->getBf() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) + " << model->getIs() << "(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    if( (con1.node->getNum() == nodeNum) && (model->getType() == NPN) )
	    outFile << " (" << model->getIs() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) - (" << model->getIs() << "/"
	    << model->getBr() << ")(1+" << model->getBr() << ")(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) "
	    << "- (" << model->getIs() << "/" << model->getBf() << ")(1+" 
	    << model->getBf() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) - " << model->getIs() << "(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    
	    if( (con0.node->getNum() == nodeNum) && (model->getType() == PNP) )
	    outFile << " (-" << model->getIs() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) + (" << model->getIs() << "/"
	    << model->getBr() << ")(1+" << model->getBr() << ")(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    if( (con2.node->getNum() == nodeNum) && (model->getType() == PNP) )
	    outFile << " (" << model->getIs() << "/" << model->getBf() << ")(1+" 
	    << model->getBf() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) + " << model->getIs() << "(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    if( (con1.node->getNum() == nodeNum) && (model->getType() == PNP) )
	    outFile << " (" << model->getIs() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) - (" << model->getIs() << "/"
	    << model->getBr() << ")(1+" << model->getBr() << ")(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) "
	    << "- (" << model->getIs() << "/" << model->getBf() << ")(1+" 
	    << model->getBf() << ")(e^(-n(X" << con2.node->getNum() << "-X" 
	    << con1.node->getNum() << ")) - 1) - " << model->getIs() << "(e^(-n(X"
	    << con0.node->getNum() << "-X" << con1.node->getNum() << ")) - 1) ";
	    
	    		
			outFile << "BJT" << getNum()
			<< " connected: Collector->Node" << getNodeNum(0)
			<< " Base->Node" << getNodeNum(1)
			<< " Emitter->Node" << getNodeNum(2) << " ." << endl
			<< "         VALUES:" << endl
			<< "                IS=" << model->getIs()
			<< "   BF=" << model->getBf() 
			<< "   BR=" << model->getBr() << endl;
			
	  break;
	  */
	case VSource:
	  if( con0.node->getNum() == nodeNum )
	    outFile << " Is" << compNum << " ";
	  else if( con1.node->getNum() == nodeNum )
	    outFile << " (-Is" << compNum << ") ";
	  /*
	    outFile << "Source" << getNum() << " value = " << getVal()
	    << " connected from Node" << getNodeNum(0)
	    << " to Node" << getNodeNum(1) << " ." << endl;
	    */
	  
	  break;

	case ISource:
	  if( con0.node->getNum() == nodeNum )
	    outFile << " (" << value << ") ";
	  else if( con1.node->getNum() == nodeNum )
	    outFile << " (-" << value << ") ";
	  break;
	  
	case Diode:
	  outFile << "Diode" << getNum() << " value = " << getVal()
	    << " connected from Node" << getNodeNum(0)
			  << " to Node" << getNodeNum(1) << " ." << endl;
			break;

	case Resistor:
	  if( con0.node->getNum() == nodeNum ){
	    outFile << " (";
	    if( con0.node->getNum() != datum )
	      outFile << "X" << con0.node->getNum();
	    if( con1.node->getNum() != datum )
	      outFile << "-X" << con1.node->getNum();
	    outFile << ")/" << value << " ";
	  }
	  if( con1.node->getNum() == nodeNum ){
	    outFile << " (";
	    if( con1.node->getNum() != datum )
	      outFile << "X" << con1.node->getNum();
	    if( con0.node->getNum() != datum )
	      outFile << "-X" << con0.node->getNum();
	    outFile << ")/" << value << " ";
	  }
	  break;
	  
	  

/*
	case Resistor:
	  if( con0.node->getNum() == nodeNum )
	    outFile << " (X" << nodeNum << "-X" << con1.node->getNum() << ")/" 
		    << value << " ";
	  else if( con1.node->getNum() == nodeNum )
	    outFile << " (X" << nodeNum << "-X" << con0.node->getNum() << ")/"
		    << value << " ";
	  
		    outFile << "Resistor" << getNum() << " value = " << getVal()
		    << " connected from Node" << getNodeNum(0)
		    << " to Node" << getNodeNum(1) << " ." << endl;
		    
	  break;
*/

	case Capacitor:
	  outFile << " 0 ";
	  /*
	    outFile << "Capacitor" << getNum() << " value = " << getVal()
	    << " connected from Node" << getNodeNum(0)
	    << " to Node" << getNodeNum(1) << " ." << endl;
	    */
	  break;
	case Inductor:
	  if( con0.node->getNum() == nodeNum )
	    outFile << " Il" << compNum << " ";
	  else if( con1.node->getNum() == nodeNum )
	    outFile << " (-Il" << compNum << ") ";

	  /*	  outFile << "Inductor" << getNum() << " value = " << getVal()
	    << " connected from Node" << getNodeNum(0)
	      << " to Node" << getNodeNum(1) << " ." << endl;
	  break;
	  */

	};
	return;
}

void Component::specialPrint( ofstream &outFile, int datum ){
	if( type == VSource ){
		outFile << " (";
		if( con0.node->getNum() != datum )
			outFile << "X" << con0.node->getNum();
		if( con1.node->getNum() != datum )
			outFile << "-X" << con1.node->getNum();
		outFile << ") -" << value << " = 0" << endl;
	}
	else if( type == Inductor ){
		outFile << " (";
		if( con0.node->getNum() != datum )
			outFile << "X" << con0.node->getNum();
		if( con1.node->getNum() != datum )
			outFile << "-X" << con1.node->getNum();
		outFile << ") = 0" << endl;
	}
}


/*
void Component::specialPrint( ofstream &outFile ){
  if( type == VSource )
    outFile << " (X" << con0.node->getNum() << "-X" << con1.node->getNum() << ") - " 
	    << value << " = 0" << endl;
}
*/


Node* Component::getNode( int conNum ){
  switch(conNum) {
  case 0:
    return con0.node;
  case 1:
    return con1.node;
  case 2:
    return con2.node;
  case 3:
    return con3.node;
  }
}

int Component::getNodeNum( int conNum ){
  switch(conNum) {
  case 0:
    return con0.node->getNum();
  case 1:
    return con1.node->getNum();
  case 2:
    return con2.node->getNum();
  case 3:
    return con3.node->getNum();
  }
}

Node::Node( int Num ){
	next = NULL;
	nodeNum = Num;
	conCount = 0;
	conList = NULL;
	}

Node::~Node(){};

int Node::getNum(){
	return nodeNum;
	}

int Node::getCount(){
	return conCount;
	}

void Node::connect( int conNumIn, Component* compIn ){
	Connections* conPtr;
	conCount++;
	if( conList == NULL ){
		conList = new Connections;
		conList->next = NULL;
		conList->conNum = conNumIn;
		conList->comp = compIn;
	}
	else {
		conPtr = conList;
		while ( conPtr->next != NULL )
			conPtr = conPtr->next;
		conPtr->next = new Connections;
		conPtr = conPtr->next;
		conPtr->next = NULL;
		conPtr->conNum = conNumIn;
		conPtr->comp = compIn;
		}
	}

Node* Node::getNext(){
	return next;
	}
void Node::setNext( Node* nodeIn ){
  next = nodeIn;
}

void Node::printComps( ofstream &outFile, int datum ){
	Connections* conPtr;
	Component* compPtr;
	conPtr = conList;
	while( conPtr->next != NULL ){
	  conPtr->comp->print( nodeNum, outFile, datum );
		outFile << "+";
		conPtr = conPtr->next;
		}
	conPtr->comp->print( nodeNum, outFile, datum );
	outFile << " = 0" << endl;
	return;
	}


NodeHead::NodeHead(){
	nodeList = NULL;
	nodeCount = 0;
	}

NodeHead::~NodeHead(){};

Node* NodeHead::addNode(){
	Node* nodePtr;
	nodeCount++;
	if( nodeList == NULL ){
		nodeList = new Node(nodeCount);
		return nodeList;
		}
	else {
		nodePtr = nodeList;
		while( nodePtr->getNext() != NULL )
			nodePtr = nodePtr->getNext();
		nodePtr->setNext( new Node(nodeCount) );
	       	return nodePtr->getNext();
		}
	}

int NodeHead::getCount(){
	return nodeCount;
	}

Node* NodeHead::getNode( int nodeNum ){
	Node* nodePtr;
	// need check that nodeNum does not exceed node count
	nodePtr = nodeList;
	for( int a=0; a < nodeNum; a++ )
		nodePtr = nodePtr->getNext();
	return nodePtr;
	}

CompHead::CompHead(){
	compList = NULL;
	mCount = 0;
	bCount = 0;
	iCount = 0;
	rCount = 0;
	dCount = 0;
	cCount = 0;
	vSCount = 0;
	iSCount = 0;
	}

CompHead::~CompHead(){};

void CompHead::addComp( Component* component ){
	Component* compPtr;
	switch ( component->getType() ){
	case ISource:
	  iSCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( iSCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( iSCount );
		}
	  break;
	case VSource:
	  vSCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( vSCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( vSCount );
		}
	  break;
	case Resistor:
	  rCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( rCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( rCount );
		}
	  break;
	case MOSFET:
	  mCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( mCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( mCount );
		}
	  break;
	case BJT:
	  bCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( bCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( bCount );
		}
	  break;
	case Diode:
	  dCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( dCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( dCount );
		}
	  break;
        case Capacitor:
	  cCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( cCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( cCount );
		}
	  break;
        case Inductor:
	  iCount++;
	  if( compList == NULL ){
	    compList = component;
	    component->setNum( iCount );
	  }
	  else{
	    compPtr = compList;
	    while( compPtr->getNext() != NULL )
	      compPtr = compPtr->getNext();
	    compPtr->setNext( component );
	    compPtr = compPtr->getNext();
	    compPtr->setNum( iCount );
		}
	  break;
	}
}

int CompHead::getCount( CompType type){
  switch (type) {
  case ISource:
    return iSCount;
  case VSource:
    return vSCount;
  case Resistor:
    return rCount;
  case Diode:
    return dCount;
  case MOSFET:
    return mCount;
  case BJT:
    return bCount;
  case Capacitor:
    return cCount;
  case Inductor:
    return iCount;
  }
}

Component* CompHead::getComp( int compNum ){
	Component* compPtr;
	compPtr = compList;
	for( int a=0; a<compNum; a++ )
		compPtr = compPtr->getNext();
	return compPtr;
	}

Model::Model( char* nameIn, TranType typeIn, double isIn, double bfIn, 
	      double brIn, double tempIn ) {
  strcpy( name, nameIn );
  type = typeIn;
  is = isIn;
  bf = bfIn;
  br = brIn;
  temp = tempIn;
  next = NULL;
}

TranType Model::getType(){
  return type;
}

char* Model::getName(){
  return name;
}

double Model::getIs(){
  return is;
}

double Model::getBf(){
  return bf;
}

double Model::getBr(){
  return br;
}

double Model::getTemp(){
  return temp;
}

void Model::setNext( Model* nextIn ){
  next = nextIn;
}

Model* Model::getNext(){
  return next;
}


ModelHead::ModelHead(){
  modelList = NULL;
}

void ModelHead::addModel( Model* modelIn ){
  Model* modelPtr;
  if( modelList == NULL ){
    modelList = modelIn;
  }
  else{
    modelPtr = modelList;
    while( modelPtr->getNext() != NULL ){
      modelPtr = modelPtr->getNext();
    }
    modelPtr->setNext( modelIn );
  }
}

Model* ModelHead::getModel ( char* nameIn ){
  Model* modelPtr = modelList;
  while( strcmp( modelPtr->getName(), nameIn ) ){
    modelPtr = modelPtr->getNext();
  }
  return modelPtr;
}





//  got to take in the component value in a way that allows for K's and M's and scientific notation.





